@@ -2,7 +2,7 @@ import asyncio
 import logging
 import warnings
 from functools import partial, update_wrapper
-from typing import (  # noqa
+from typing import (
     TYPE_CHECKING,
     Any,
     AsyncIterator,
@@ -22,6 +22,9 @@ from typing import (  # noqa
     cast,
 )
 
+from aiosignal import Signal
+from frozenlist import FrozenList
+
 from . import hdrs
 from .abc import (
     AbstractAccessLogger,
@@ -29,11 +32,9 @@ from .abc import (
     AbstractRouter,
     AbstractStreamWriter,
 )
-from .frozenlist import FrozenList
 from .helpers import DEBUG
 from .http_parser import RawRequestMessage
 from .log import web_logger
-from .signals import Signal
 from .streams import StreamReader
 from .web_log import AccessLogger
 from .web_middlewares import _fix_request_current_app
@@ -56,12 +57,13 @@ __all__ = ("Application", "CleanupError")
 
 
 if TYPE_CHECKING:  # pragma: no cover
+    from .typedefs import Handler
+
     _AppSignal = Signal[Callable[["Application"], Awaitable[None]]]
     _RespPrepareSignal = Signal[Callable[[Request, StreamResponse], Awaitable[None]]]
-    _Handler = Callable[[Request], Awaitable[StreamResponse]]
     _Middleware = Union[
-        Callable[[Request, _Handler], Awaitable[StreamResponse]],
-        Callable[["Application", _Handler], Awaitable[_Handler]],  # old-style
+        Callable[[Request, Handler], Awaitable[StreamResponse]],
+        Callable[["Application", Handler], Awaitable[Handler]],  # old-style
     ]
     _Middlewares = FrozenList[_Middleware]
     _MiddlewaresHandlers = Optional[Sequence[Tuple[_Middleware, bool]]]
@@ -70,7 +72,6 @@ else:
     # No type checker mode, skip types
     _AppSignal = Signal
     _RespPrepareSignal = Signal
-    _Handler = Callable
     _Middleware = Callable
     _Middlewares = FrozenList
     _MiddlewaresHandlers = Optional[Sequence]
@@ -110,7 +111,7 @@ class Application(MutableMapping[str, Any]):
         handler_args: Optional[Mapping[str, Any]] = None,
         client_max_size: int = 1024 ** 2,
         loop: Optional[asyncio.AbstractEventLoop] = None,
-        debug: Any = ...  # mypy doesn't support ellipsis
+        debug: Any = ...,  # mypy doesn't support ellipsis
     ) -> None:
         if router is None:
             router = UrlDispatcher()
@@ -278,7 +279,7 @@ class Application(MutableMapping[str, Any]):
     @property
     def debug(self) -> bool:
         warnings.warn("debug property is deprecated", DeprecationWarning, stacklevel=2)
-        return self._debug
+        return self._debug  # type: ignore[no-any-return]
 
     def _reg_subapp_signals(self, subapp: "Application") -> None:
         def reg_handler(signame: str) -> None:
@@ -365,7 +366,7 @@ class Application(MutableMapping[str, Any]):
         *,
         loop: Optional[asyncio.AbstractEventLoop] = None,
         access_log_class: Type[AbstractAccessLogger] = AccessLogger,
-        **kwargs: Any
+        **kwargs: Any,
     ) -> Server:
 
         if not issubclass(access_log_class, AbstractAccessLogger):
@@ -384,10 +385,10 @@ class Application(MutableMapping[str, Any]):
                 kwargs[k] = v
 
         return Server(
-            self._handle,  # type: ignore
+            self._handle,  # type: ignore[arg-type]
             request_factory=self._make_request,
             loop=self._loop,
-            **kwargs
+            **kwargs,
         )
 
     def make_handler(
@@ -395,7 +396,7 @@ class Application(MutableMapping[str, Any]):
         *,
         loop: Optional[asyncio.AbstractEventLoop] = None,
         access_log_class: Type[AbstractAccessLogger] = AccessLogger,
-        **kwargs: Any
+        **kwargs: Any,
     ) -> Server:
 
         warnings.warn(
@@ -427,7 +428,11 @@ class Application(MutableMapping[str, Any]):
 
         Should be called after shutdown()
         """
-        await self.on_cleanup.send(self)
+        if self.on_cleanup.frozen:
+            await self.on_cleanup.send(self)
+        else:
+            # If an exception occurs in startup, ensure cleanup contexts are completed.
+            await self._cleanup_ctx._on_cleanup(self)
 
     def _make_request(
         self,
@@ -477,7 +482,7 @@ class Application(MutableMapping[str, Any]):
         match_info.freeze()
 
         resp = None
-        request._match_info = match_info  # type: ignore
+        request._match_info = match_info
         expect = request.headers.get(hdrs.EXPECT)
         if expect:
             resp = await match_info.expect_handler(request)
@@ -488,13 +493,13 @@ class Application(MutableMapping[str, Any]):
 
             if self._run_middlewares:
                 for app in match_info.apps[::-1]:
-                    for m, new_style in app._middlewares_handlers:  # type: ignore  # noqa
+                    for m, new_style in app._middlewares_handlers:  # type: ignore[union-attr] # noqa
                         if new_style:
                             handler = update_wrapper(
                                 partial(m, handler=handler), handler
                             )
                         else:
-                            handler = await m(app, handler)  # type: ignore
+                            handler = await m(app, handler)  # type: ignore[arg-type]
 
             resp = await handler(request)
 
@@ -514,7 +519,7 @@ class Application(MutableMapping[str, Any]):
 class CleanupError(RuntimeError):
     @property
     def exceptions(self) -> List[BaseException]:
-        return self.args[1]
+        return cast(List[BaseException], self.args[1])
 
 
 if TYPE_CHECKING:  # pragma: no cover
@@ -544,7 +549,7 @@ class CleanupContext(_CleanupContextBase):
             except Exception as exc:
                 errors.append(exc)
             else:
-                errors.append(RuntimeError("{!r} has more than one 'yield'".format(it)))
+                errors.append(RuntimeError(f"{it!r} has more than one 'yield'"))
         if errors:
             if len(errors) == 1:
                 raise errors[0]
